(*
Copyright 2009-2013 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Conditional_Rewriting
imports
  Trs
  Reduction_Pair

begin

type_synonym ('f, 'v) condition = "('f, 'v) term \<times> ('f, 'v) term"
type_synonym ('f, 'v) crule = "('f, 'v) rule \<times> ('f, 'v) condition list"
type_synonym ('f, 'v) ctrs = "('f, 'v) crule set"

fun cstep_n :: "('f, 'v) ctrs \<Rightarrow> nat \<Rightarrow> ('f, 'v) term rel"
where
  "cstep_n R 0 = {}" |
  "cstep_n R (Suc n) =
    {(C\<langle>l \<cdot> \<sigma>\<rangle>, C\<langle>r \<cdot> \<sigma>\<rangle>) | C l r \<sigma> cs.
      ((l, r), cs) \<in> R \<and> (\<forall> (s\<^sub>i, t\<^sub>i) \<in> set cs. (s\<^sub>i \<cdot> \<sigma>, t\<^sub>i \<cdot> \<sigma>) \<in> (cstep_n R n)\<^sup>*)}"

lemma cstep_n_SucI:
  assumes "((l, r), cs) \<in> R"
    and "\<forall>(s\<^sub>i, t\<^sub>i) \<in> set cs. (s\<^sub>i \<cdot> \<sigma>, t\<^sub>i \<cdot> \<sigma>) \<in> (cstep_n R n)\<^sup>*"
    and "s = C\<langle>l \<cdot> \<sigma>\<rangle>"
    and "t = C\<langle>r \<cdot> \<sigma>\<rangle>"
  shows "(s, t) \<in> cstep_n R (Suc n)"
  using assms by (auto) blast

lemma cstep_n_SucE:
  assumes "(s, t) \<in> cstep_n R (Suc n)"
  obtains C l r \<sigma> cs where "((l, r), cs) \<in> R"
    and "\<forall>(s\<^sub>i, t\<^sub>i) \<in> set cs. (s\<^sub>i \<cdot> \<sigma>, t\<^sub>i \<cdot> \<sigma>) \<in> (cstep_n R n)\<^sup>*"
    and "s = C\<langle>l \<cdot> \<sigma>\<rangle>"
    and "t = C\<langle>r \<cdot> \<sigma>\<rangle>"
  using assms by auto
declare cstep_n.simps(2) [simp del]

lemma cstep_n_Suc_mono:
  "cstep_n R n \<subseteq> cstep_n R (Suc n)"
proof (induct n)
  case (Suc n)
  show ?case
    using rtrancl_mono [OF Suc] by (auto elim!: cstep_n_SucE intro!: cstep_n_SucI)
qed simp

lemma cstep_n_mono:
  assumes "i \<le> n"
  shows "cstep_n R i \<subseteq> cstep_n R n"
using assms
proof (induct "n - i" arbitrary: i)
  case (Suc k)
  show ?case
  proof (cases "i = n")
    assume "i \<noteq> n"
    with Suc(1) [of "Suc i"] and Suc(2-)
      have "cstep_n R (Suc i) \<subseteq> cstep_n R n" by arith
    with cstep_n_Suc_mono [of R i]
      show ?case by blast
  qed simp
qed simp

lemma cstep_n_ctxt_subst:
  assumes step: "(s,t) \<in> cstep_n R n"
  shows "(D \<langle>s \<cdot> \<sigma> \<rangle>,D \<langle>t \<cdot> \<sigma> \<rangle>) \<in> cstep_n R n"
  using step
proof (induct n arbitrary: s t D)
  case 0
  thus ?case by simp
next
  case (Suc n s t D)
  from Suc(2) obtain l r cs C \<mu> where lr: "((l,r),cs) \<in> R" and
    cs: "\<And> si ti.  (si,ti) \<in> set cs \<Longrightarrow> (si \<cdot> \<mu>, ti \<cdot> \<mu>) \<in> (cstep_n R n)^*"
    and s: "s = C \<langle> l \<cdot> \<mu> \<rangle>" and t: "t = C \<langle> r \<cdot> \<mu> \<rangle>" by (blast elim: cstep_n_SucE)
  show ?case unfolding s t
  proof (rule cstep_n_SucI [OF lr, of "\<mu> \<circ>s \<sigma>" _ _ "D \<circ>c (C \<cdot>c \<sigma>)"], rule ballI2)
    fix si ti 
    {
      fix s t
      assume "(s,t) \<in> (cstep_n R n)^*"
      hence "(s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> (cstep_n R n)^*"
      proof (induct)
        case base show ?case by simp
      next
        case (step t u)
        from step(3) Suc(1)[OF step(2), of Hole] show ?case by auto
      qed
    } note main = this
    assume "(si, ti) \<in> set cs" 
    from main[OF cs[OF this]]
    show "(si \<cdot> \<mu> \<circ>s \<sigma>, ti \<cdot> \<mu> \<circ>s \<sigma>) \<in> (cstep_n R n)^*" by simp
  qed auto
qed

definition "cstep R \<equiv> \<Union>n. cstep_n R n"

lemma cstep_n_mono':
  "cstep_n R m \<subseteq> cstep_n R (m + k)"
  using cstep_n_mono [of "m" "m + k"] by simp

lemma csteps_n_trans:
 assumes "(s,t) \<in> (cstep_n R m)^+" and "(t,u) \<in> (cstep_n R k)^*"
 shows "\<exists> n. (s,u) \<in> (cstep_n R n)^+"
proof- 
  let ?n = "max m k"
  from max.cobounded1[of m k, unfolded Nat.le_iff_add] obtain m' where "?n = m + m'" by blast
  with trancl_mono[OF assms(1) cstep_n_mono] have 1:"(s,t) \<in> (cstep_n R ?n)^+" by auto
  from max.cobounded2[of k m, unfolded Nat.le_iff_add] obtain k' where "?n = k + k'" by blast 
  with assms(2)  rtrancl_mono [OF cstep_n_mono'] 1
    have 2:"(t,u) \<in> (cstep_n R ?n)^*" by (auto)
  show ?thesis by (rule exI[of _ ?n], insert 1 2, auto)
qed

lemma csteps_imp_csteps_n:
 assumes "(s,t) \<in> (cstep R)^*" 
 shows "\<exists> n. (s,t) \<in> (cstep_n R n)^*"
proof- 
 from rtrancl_imp_seq[OF assms] obtain ti k where 
  ti:"ti 0 = s" "ti k = t" "(\<forall>i<k. (ti i, ti (Suc i)) \<in> cstep R)" by blast
 from this(3)[unfolded cstep_def] have "\<forall> i < k. \<exists> ni. (ti i,ti (Suc i)) \<in> cstep_n R ni" by blast
 then obtain ni where ni:"\<forall> i < k. (ti i,ti (Suc i)) \<in> cstep_n R (ni i)" by metis 
 let ?nis = "set (map ni [0..<k])" let ?n = "Max ?nis"
 have "\<forall>i < k. (ti i,ti (Suc i)) \<in> cstep_n R ?n" 
  proof(rule,rule)
   fix i
   assume i:"i < k"
   with Max_ge[of ?nis "ni i"] have "ni i \<le> ?n" by fastforce
   then obtain k where n:"?n = ni i + k" by (metis Nat.le_iff_add)
   from ni[rule_format, OF i] cstep_n_mono' show "(ti i,ti (Suc i)) \<in> cstep_n R ?n"
     unfolding n by auto
 qed
 with ti(1) ti(2) relpow_fun_conv[of s t k "cstep_n R ?n"] have "(s,t) \<in> (cstep_n R ?n) ^^ k" by blast
 from relpow_imp_rtrancl[OF this] show ?thesis by blast
qed

lemma all_cstep_imp_cstep_n:
 assumes "\<forall> i < length cs. (si i,ti i) \<in> (cstep R)^*" shows "\<exists> n.(\<forall>i < length cs. (si i,ti i) \<in> (cstep_n R n)^*)" 
proof-
 from csteps_imp_csteps_n assms have "\<forall> i < length cs. \<exists> ni. (si i,ti i) \<in> (cstep_n R ni)^*" by blast
 then obtain ni where ni:"\<forall> i < length cs. (si i,ti i) \<in> (cstep_n R (ni i))^*" by metis
 let ?nis = "set (map ni [0..<length cs])" let ?n = "Max ?nis"
 have "\<forall>i < length cs. (si i,ti i) \<in> (cstep_n R ?n)^*" 
  proof(rule,rule)
   fix i
   assume i:"i < length cs"
   with Max_ge[of ?nis "ni i"] have "ni i \<le> ?n" by fastforce
   then obtain k where n:"?n = ni i + k" by (metis Nat.le_iff_add)
   from ni[rule_format, OF i] rtrancl_mono[OF cstep_n_mono'] show "(si i, ti i) \<in> (cstep_n R ?n)^*"
     unfolding n by blast
 qed
 thus ?thesis by auto
qed

lemma cstep_ctxt_closed:"ctxt_closed (cstep R)"
 proof
  fix s t C
  assume "(s,t) \<in> cstep R" 
  then obtain n where "(s,t) \<in> cstep_n R n" unfolding cstep_def by auto 
  from cstep_n_ctxt_subst[OF this, of C Var] show "(C\<langle>s\<rangle>, C\<langle>t\<rangle>) \<in> cstep R" 
   unfolding cstep_def subst_apply_term_empty by fast
 qed

definition quasi_reductive :: "('f,'v)ctrs \<Rightarrow> bool"
  where "quasi_reductive R \<equiv> \<exists> S NS. redpair_order S NS \<and> ctxt_closed S \<and> (\<forall> l r cs (\<sigma> :: ('f,'v)subst). ((l,r),cs) \<in> R \<longrightarrow> 
         (\<forall> i. i < length cs 
         \<longrightarrow> (\<forall> j. j < i \<longrightarrow> (fst (cs ! j) \<cdot> \<sigma>, snd (cs ! j) \<cdot> \<sigma>) \<in> S^=) 
         \<longrightarrow> (l \<cdot> \<sigma>, fst (cs ! i) \<cdot> \<sigma>) \<in> (S \<union> {\<rhd>})^+) \<and> 
         ((\<forall> j. j < length cs \<longrightarrow> (fst (cs ! j) \<cdot> \<sigma>, snd (cs ! j) \<cdot> \<sigma>) \<in> S^=)
          \<longrightarrow> (l \<cdot> \<sigma>, r \<cdot> \<sigma>) \<in> S)
        )"

lemma quasi_reductive_SN: assumes qr: "quasi_reductive (R :: ('f,'v)ctrs)"
  shows "SN (cstep R)"
proof -
  from qr[unfolded quasi_reductive_def]
  obtain S NS where rp: "redpair_order S NS"
  and mono: "ctxt_closed S"
  and qr: "\<And> l r cs (\<sigma> :: ('f,'v)subst).
             ((l, r), cs) \<in> R \<Longrightarrow>
             (\<forall>j<length cs. (fst (cs ! j) \<cdot> \<sigma>, snd (cs ! j) \<cdot> \<sigma>) \<in> S^=) \<Longrightarrow> (l \<cdot> \<sigma>, r \<cdot> \<sigma>) \<in> S" by blast
  interpret redpair_order S NS by fact
  show ?thesis
  proof (rule SN_subset[OF SN])
    show "cstep R \<subseteq> S"
    proof
      fix s t
      assume "(s,t) \<in> cstep R"
      then obtain n where "(s,t) \<in> cstep_n R n" unfolding cstep_def by auto
      thus "(s,t) \<in> S"
      proof (induct n arbitrary: s t)
        case 0 thus ?case by simp
      next
        case (Suc n s t)
        from Suc(2)
        obtain l r cs C \<sigma> where lr: "((l,r),cs) \<in> R" and
          cs: "\<And> si ti.  (si,ti) \<in> set cs \<Longrightarrow> (si \<cdot> \<sigma>, ti \<cdot> \<sigma>) \<in> (cstep_n R n)^*"
          and s: "s = C \<langle> l \<cdot> \<sigma> \<rangle>" and t: "t = C \<langle> r \<cdot> \<sigma> \<rangle>" by (blast elim: cstep_n_SucE)
        note qr = qr[OF lr, of \<sigma>]
        have "(l \<cdot> \<sigma>, r \<cdot> \<sigma>) \<in> S"
        proof (rule qr[rule_format])
          fix j
          assume j: "j < length cs"
          obtain sj tj where csj: "cs ! j = (sj,tj)" by force
          from j have "(sj,tj) \<in> set cs" unfolding csj[symmetric] set_conv_nth by auto
          from cs[OF this] have steps: "(sj \<cdot> \<sigma>, tj \<cdot> \<sigma>) \<in> (cstep_n R n)^*" .
          {
            fix s t
            assume "(s,t) \<in> (cstep_n R n)^*"
            hence "(s,t) \<in> S^="
            proof (induct)
              case base thus ?case by simp
            next
              case (step t u)
              from step(3) Suc(1)[OF step(2)] have "(s,u) \<in> S^= O S" by auto
              thus "(s,u) \<in> S^=" using trans_S_point by blast
            qed
          }
          from this[OF steps]
          show "(fst (cs ! j) \<cdot> \<sigma>, snd (cs ! j) \<cdot> \<sigma>) \<in> S^=" unfolding csj by simp
        qed
        from ctxt_closedE[OF mono this, of C]
        show ?case unfolding s t .
      qed
    qed
  qed
qed

end

