(*
Copyright 2009-2013 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Innermost_Impl
imports
  Q_Restricted_Rewriting
  Matching
  Trs_Impl
begin

definition
  NF_terms_list :: "('f, 'v) term list \<Rightarrow> ('f, 'w) term \<Rightarrow> bool"
where
  "NF_terms_list ls t = (\<forall>s\<in>set (supteq_list t). \<forall>l\<in>set ls. \<not> matches s l)"

lemma inj_NF_terms_list_sound:
  assumes inj: "inj f"
  shows "NF_terms_list Q t \<longleftrightarrow> t \<in> NF_terms (map_vars_term f ` set Q)"
    (is "_ \<longleftrightarrow> t \<in> NF_terms ?Q")
proof
  assume "NF_terms_list Q t"
  show "t \<in> NF_terms ?Q"
  proof (rule ccontr)
    assume "t \<notin> NF_terms ?Q"
    then obtain u where "(t, u) \<in> rstep (Id_on ?Q)" by auto
    then obtain C l r \<sigma> where "(l, r) \<in> Id_on ?Q" and t: "t = C\<langle>l \<cdot> \<sigma>\<rangle>" by best
    then obtain l' where l: "l = map_vars_term f l'" and "l' \<in> set Q" by blast
    hence id: "(l', l') \<in> Id_on (set Q)" by auto
    from t have subt: "t \<unrhd> l \<cdot> \<sigma>" by auto
    from `NF_terms_list Q t`
      have "\<forall>s\<unlhd>t. \<forall>(l, r)\<in>Id_on (set Q). \<not> matches s l"
      unfolding NF_terms_list_def by auto
    with id have "\<forall>s\<unlhd>t. \<not> matches s l'" by blast
    with subt have "\<not> matches (l \<cdot> \<sigma>) l'" by simp
    from this[unfolded l] show False by (auto simp: map_subst_domain matches_iff)
  qed
next
  assume NF: "t \<in> NF_terms ?Q"
  show "NF_terms_list Q t"
  proof (rule ccontr)
    assume "\<not> NF_terms_list Q t"
    then obtain s l \<sigma> where "t \<unrhd> s" and "l \<in> set Q" and s: "s = l \<cdot> \<sigma>"
      unfolding NF_terms_list_def by (auto simp: matches_iff)
    from `t \<unrhd> s` obtain C where t: "t = C\<langle>s\<rangle>" by auto
    let ?l = "map_vars_term f l"
    let ?\<sigma> = "map_subst_domain (the_inv f) \<sigma>"
    from `l \<in> set Q` have "(?l, ?l) \<in> Id_on ?Q" by auto
    hence "(?l \<cdot> ?\<sigma>, ?l \<cdot> ?\<sigma>) \<in> rstep (Id_on ?Q)" by auto
    moreover have "?l \<cdot> ?\<sigma> = l \<cdot> \<sigma>" unfolding map_subst_domain[symmetric]
      unfolding map_vars_term_compose o_def the_inv_f_f[OF inj] by (simp add: term.map_ident)
    ultimately have "(t, t) \<in> rstep (Id_on ?Q)" unfolding s t by auto
    with NF show False by auto
  qed
qed

lemma map_vars_id': "map_vars_term id = id" by (rule ext) simp

lemma NF_terms_list_sound[simp]:
  "NF_terms_list Q t \<longleftrightarrow> t \<in> NF_terms (set Q)"
   using inj_NF_terms_list_sound[of id] unfolding map_vars_id' by auto

text {*Partially applied version of @{thm NF_terms_list_sound}.*}
lemma NF_terms_list_sound': "NF_terms_list Q = (\<lambda>t. t \<in> NF_terms (set Q))"
  by (rule ext) (simp only: NF_terms_list_sound)

definition check_NF_terms_subset :: "('f,'v)term list \<Rightarrow> ('f,'v)term list \<Rightarrow> ('f,'v)term check"
  where "check_NF_terms_subset Q Q' \<equiv> check_all (\<lambda> q. \<not> NF_terms_list Q q) Q'"

lemma check_NF_terms_subset[simp]: "isOK(check_NF_terms_subset Q Q') = (NF_terms (set Q) \<subseteq> NF_terms (set Q'))" unfolding check_NF_terms_subset_def 
  unfolding NF_terms_subset_criterion[symmetric]
  by auto

definition check_NF_terms_eq :: "('f,'v)term list \<Rightarrow> ('f,'v)term list \<Rightarrow> ('f,'v)term check"
  where "check_NF_terms_eq Q Q' \<equiv> do {
      check_NF_terms_subset Q Q';
      check_NF_terms_subset Q' Q
    }"

lemma check_NF_terms_eq[simp]: "isOK(check_NF_terms_eq Q Q') = (NF_terms (set Q) = NF_terms (set Q'))" unfolding check_NF_terms_eq_def  by auto

text {*Partially applied version of @{thm inj_NF_terms_list_sound}.*}
lemma inj_NF_terms_list_sound':
  "inj f \<Longrightarrow> NF_terms_list Q = (\<lambda>t. t \<in> NF_terms (map_vars_term f ` set Q))"
  by (rule ext) (simp only: inj_NF_terms_list_sound)

lemma NF_terms_list_Inl_sound:
  "NF_terms_list Q = (\<lambda>t. t \<in> NF_terms (map_vars_term Inl ` set Q))"
  by (simp add: inj_NF_terms_list_sound')

fun
  applicable_rule_list :: "('f, 'v) term list \<Rightarrow> ('f, 'v) rule \<Rightarrow> bool"
where
  "applicable_rule_list Q (Var _, _) = True" |
  "applicable_rule_list Q (Fun f ss, _) = (\<forall>s\<in>set ss. NF_terms_list Q s)"

lemma applicable_rule_list_sound[simp]: "applicable_rule_list Q = applicable_rule (set Q)"
proof (rule ext, unfold split_paired_all)
  fix l r show "applicable_rule_list Q (l, r) = applicable_rule (set Q) (l, r)"
  proof (cases l)
    case (Var x) show ?thesis unfolding Var applicable_rule_def by auto
  next
    case (Fun f ts) show ?thesis
      by (simp add: NF_rstep_supt_args_conv applicable_rule_def Fun)
  qed
qed

definition
  applicable_rules_list :: "('f, 'v) term list \<Rightarrow> ('f, 'v) rule list  \<Rightarrow> ('f, 'v) rules"
where
  "applicable_rules_list Q R = filter (applicable_rule_list Q) R"

lemma applicable_rules_list_sound[simp]:
  "set (applicable_rules_list Q R) = applicable_rules (set Q) (set R)"
  unfolding applicable_rules_def applicable_rules_list_def by auto

end

