theory Status_Impl
imports
  Term_Order_Impl
  Map_Choice
begin
type_synonym 'f status_impl = "(('f \<times> nat) \<times> nat list)list"

typedef 'f status_impl_check = "{ (\<sigma> :: 'f \<times> nat \<Rightarrow> nat list, b :: bool). b \<longrightarrow> 
  (\<forall> f k. set (\<sigma> (f,k)) \<subseteq> {0 ..< k})}"
  by auto

setup_lifting type_definition_status_impl_check

lift_definition status_impl_check_test :: "'f status_impl_check \<Rightarrow> bool" is snd .
lift_definition status_impl_check_status :: "'f status_impl_check \<Rightarrow> 'f status" is 
  "\<lambda> (\<sigma>,b). if b then \<sigma> else (\<lambda> _. [])" by (force split: if_splits)

lift_definition status_impl_check :: "'f status_impl \<Rightarrow> 'f status_impl_check" is
  "\<lambda> s. (fun_of_map_fun (map_of s) (\<lambda> (f,n). [0 ..< n]),
              (\<forall> fidx \<in> set s. (\<forall> i \<in> set (snd fidx). i < snd (fst fidx))))" 
proof (transfer, rule, unfold split, intro allI impI) 
  fix s :: "'f status_impl" and f n
  let ?s = "fun_of_map_fun (map_of s) (\<lambda> (f,n). [0 ..< n])"
  assume *: "\<forall>fidx\<in>set s. \<forall>i\<in>set (snd fidx). i < snd (fst fidx)"
  show "set (?s (f, n)) \<subseteq> {0..<n}"
  proof (cases "map_of s (f,n)")
    case (Some idx)
    from map_of_SomeD[OF Some] Some
    have "?s (f,n) = idx" "((f, n), idx) \<in> set s" by auto
    with * Some show ?thesis by auto
  qed auto
qed

definition status_of :: "'f status_impl \<Rightarrow> 'f status option" where
  "status_of s \<equiv> let sc = status_impl_check s
    in if status_impl_check_test sc then
      Some (status_impl_check_status sc) else None"

lemma status_of_status: assumes s: "status_of s = Some \<sigma>"
  shows "status \<sigma> = fun_of_map_fun (map_of s) (\<lambda> (f,n). [0 ..< n])"
proof -
  let ?sc = "status_impl_check s"
  from s have c: "status_impl_check_test ?sc" and \<sigma>: "\<sigma> = status_impl_check_status ?sc"
    unfolding status_of_def by (auto split: if_splits simp: Let_def)
  show ?thesis unfolding \<sigma> using c
    by (transfer, auto)
qed

hide_const 
  status_impl_check 
  status_impl_check_test 
  status_impl_check_status 

lemma status_of_default: assumes s: "status_of s = Some \<sigma>"
  and f: "(f,n) \<notin> fst ` set s"
  shows "status \<sigma> (f,n) = [0 ..< n]"
proof -
  from map_of_eq_None_iff[of s "(f,n)"] f have "map_of s (f,n) = None" by simp
  thus ?thesis unfolding status_of_status[OF s] by simp
qed
end
